#include <cmath>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_errno.h>
#include <cuba.h>
#include "debugResummation.hh"
#include "chebyshev.hh"
#include "QCD_EW.hh"
#include "math/special_functions.hh"
#include "math/integration.hh"
#include "math/struct_definitions.hh"
#include "math/mellin_Dk.hh"


using namespace std;


namespace resumm {

  extern double Lum(double x, void *p);

  debugResummation::~debugResummation() {
    return;
  }

  dcomplex debugResummation::Ug0expS_matched(dcomplex N) {
    N = NofN(N);
    return g0expS_matched(N);
  }
  dcomplex debugResummation::ULuminosityN(dcomplex N) {
    return LuminosityN(N);
  }
  double debugResummation::Lumintau(double m, double sqrt_s) {
    doRapidityDistr = false;
    tau = m*m/sqrt_s/sqrt_s;
    rapY = 1e300;
    M   = m;
    muF = m;
    muR = m;
    return Lum(tau, this);
  }
  double debugResummation::Lumintau() {
    return luminosity_in_tau;
  }
  double debugResummation::LuminxDIS(double x, double Q, double y, double TM, double mcoQ) {
    nf = nf_of_mu(Q);
    DISy = y;
    doRapidityDistr = false;
    tau = x;
    rapY = 1e300;
    M   = Q;
    muF = Q;
    muR = Q;
    TargetMass = TM;
    double TMtau = sqrt(1+4*x*x*TM*TM/Q/Q);
    double pref = (1+TMtau)/2./pow(TMtau,3);
    //
    double rescale = 1.+mcoQ*mcoQ;  // restore massive kinematics!!
    x *= rescale;
    return x * pref * Lum(x, this);
  }





  /////////////////////////////////
  //  Fixed-Order approximation  //
  /////////////////////////////////

  dcomplex debugResummation::g0expS_expanded_prescale(dcomplex N, order o) {
    dcomplex NN = NofN(N);
    dcomplex res = 0;
    if(o == LO)    res = 1;
    if(o == NLO)   res = g0expS_order1(NN);
    if(o == NNLO)  res = g0expS_order2(NN);
    if(o == NNNLO) res = g0expS_order3(NN);
    return res + (scaledep ? muFdep(NN,o) : 0);
  }
  dcomplex debugResummation::g0expS_expanded_predamping(dcomplex N, order o) {
    if(!scaledep) return g0expS_expanded_prescale(N, o);
    double lrf = 2.*log(muRbkp/muFbkp);
    double bb1 = b1*b0;
    double bb2 = b2*b0;
    dcomplex lo   = g0expS_expanded_prescale(N, LO);
    dcomplex nlo  = g0expS_expanded_prescale(N, NLO);
    dcomplex nnlo = g0expS_expanded_prescale(N, NNLO);
    dcomplex mur_dep_term = 0;
    if(o == NLO)   mur_dep_term = as*2*lo*b0*lrf;
    if(o == NNLO)  mur_dep_term = as*(as*2*bb1*lo + 3*b0*nlo) *lrf + as*as*3*b0*b0*lo*lrf*lrf;
    if(o == NNNLO) mur_dep_term = as*(as*as*2*bb2*lo + as*3*bb1*nlo + 4*b0*nnlo) *lrf
		     + as*as*(as*7*b0*bb1*lo + 6*b0*b0*nlo) *lrf*lrf + as*as*as*4*b0*b0*b0*lo*lrf*lrf*lrf;
    return g0expS_expanded_prescale(N, o) + mur_dep_term;
  }
  dcomplex debugResummation::g0expS_expanded_presubtraction(dcomplex N, order o) {
    if(DampingOrder > 0) {
      dcomplex res = 0;
      for(unsigned int i=1; i<=DampingOrder; i++) {
	res += pow(-1.,i+1) * gsl_sf_choose(DampingOrder,i) * g0expS_expanded_predamping(N+1.*i,o);
      }
      return res;
    }
    return g0expS_expanded_predamping(N,o);
  }
  dcomplex debugResummation::g0expS_expanded(dcomplex N, order o) {
    if(doNsubtraction) {
      return g0expS_expanded_presubtraction(N,o) - g0expS_expanded_presubtraction(Nsubtraction,o);
    }
    return g0expS_expanded_presubtraction(N,o);
  }
  dcomplex debugResummation::lum_times_g0expS_expanded(dcomplex N, order o) {
    return g0expS_expanded(N,o) * LuminosityN(N);
  }
  double FOapprox_integrand(double t, void *p) {
    debugResummation *Mres = (debugResummation *) p;
    double x = Mres->tau;
    if(Mres->proc < 0) {
      double tau = sqrt(1+4*x*x*Mres->TargetMass*Mres->TargetMass/Mres->M/Mres->M);
      x *= 2./(1.+tau);
    }
    dcomplex N = Mres->MP_reN0 + (Mres->MP_slope-I)*log(t);
    dcomplex res = (I-Mres->MP_slope) * exp(-N*log(x)) * Mres->lum_times_g0expS_expanded(N,Mres->matching_order)/t;
    return imag(res)/M_PI;
  }
  double debugResummation::FOapprox() {
    double res=0, prec = 1.e-3;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(FOapprox_integrand,0.,1.,prec,NULL,this);
    gsl_set_error_handler (old_handler);
    return res;
  }
  double debugResummation::FOapprox_xs(double m, double sqrt_s,
				double muF_m_ratio, double muR_m_ratio,
				log_order LOGo, order match_order, double *asinput) {
    doRapidityDistr = false;
    //
    tau = m*m/sqrt_s/sqrt_s;
    rapY = 1e300;
    M   = m;
    muF = m*muF_m_ratio;
    muR = m*muR_m_ratio;
    muFbkp = muF;
    muRbkp = muR;
    //
    LL_order = LOGo;
    matching_order = match_order;
    //
    if(asinput == NULL) as = alphas(muR); // /50.;
    else as = *asinput; // /50.;
    //cout << "  " << muR << "  "<< as << endl;
    //
    bool redo_cheb = forceredo||!(last_Y==rapY && last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    if(scaledep) { muF=M; muR=M; }
    ComputeCommon(redo_cheb);
    if(scaledep) { muF = m*muF_m_ratio;  muR = m*muR_m_ratio; luminosity_in_tau = Lum(tau, this); }
    Chebyshev cheb(this);
    //
    double result = 0.;
    //
    MP_slope = 1.;
    MP_reN0 = MPbeta + 3.;
    double NL = Nlandau();
    if(MP_reN0 >= NL) {
      MP_reN0 = ( MPbeta + NL )/2.;
      if(MP_reN0 >= NL) {
	cout << "ERROR: Landau pole N_L = " << NL
	     << " to the left of the rightmost singularity: " << MPbeta << endl;
	exit(-13);
      }
    }
    MPzmin=0.8*tau*(1-tau/2.);
    //if(tau<MPzmin) MPzmin=tau;
    if(tau>0.5) MPzmin=0.3;
    //
    if(redo_cheb) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
    //
    if(scaledep) { muF=M; muR=M; }
    result = FOapprox()/luminosity_in_tau;
    //
    last_muFrat = muF_m_ratio;
    last_muRrat = muR_m_ratio;
    last_M      = M;
    last_Y      = rapY;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = Hmtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit = false;
    //
    return result;
  }
  double debugResummation::FOapprox_xs_rap(double m, double Y, double sqrt_s,
				    double muF_m_ratio, double muR_m_ratio,
				    log_order LOGo, order match_order, double *asinput) {
    doRapidityDistr = true;
    //
    rapY = Y;
    tau = m*m/sqrt_s/sqrt_s;
    M   = m;
    muF = m*muF_m_ratio;
    muR = m*muR_m_ratio;
    //
    LL_order = LOGo;
    matching_order = match_order;
    //
    if(asinput == NULL) as = alphas(muR); // /50.;
    else as = *asinput; // /50.;
    //cout << as << endl;
    //
    bool redo_cheb = forceredo||!(last_Y==rapY && last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    ComputeCommon(redo_cheb);
    Chebyshev cheb(this);
    //
    double result = 0.;
    //
    MP_slope = 1.;
    MP_reN0 = MPbeta + 5.;
    double NL = Nlandau();
    if(MP_reN0 >= NL) {
      MP_reN0 = ( MPbeta + NL )/2.;
      if(MP_reN0 >= NL) {
	cout << "ERROR: Landau pole N_L = " << NL
	     << " to the left of the rightmost singularity: " << MPbeta << endl;
	exit(-13);
      }
    }
    MPzmin=0.8*tau*(1-tau/2.);
    if(tau>0.5) MPzmin=0.3;
    //MPzmin = tau*exp(2*fabs(Y)); // we can use this tighter lower limit
    //if(tau<MPzmin) MPzmin=tau;
    //if(tau>0.01) MPzmin = 0.01;
    //double Ythr = -0.5*log(MPzmin) * 0.7;
    //double Yeff = (fabs(Y) < Ythr ? Y : Y/fabs(Y)*Ythr);
    //MPzmin *= exp(2*fabs(Yeff)); // rescale?
    if(redo_cheb) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
    result = FOapprox()/luminosity_in_tau;
    //
    last_muFrat = muF_m_ratio;
    last_muRrat = muR_m_ratio;
    last_M      = M;
    last_Y      = rapY;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = Hmtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit = false;
    //
    return result;
  }










  struct s_lum_Mellin  { Resummation *res; double N;};
  
  double Lum_Mellin_integrand(double z, void* par)
  {
    s_lum_Mellin pars = *(s_lum_Mellin *) par;
    Resummation *res = pars.res;
    double N   = pars.N;
    double lum = Lum(z,res);
    return lum*pow(z,N-1.);
  }

  double debugResummation::LumN(double N)
  {
    s_lum_Mellin pars = {this, N};
    double result = integration::gauss(Lum_Mellin_integrand,1e-5,1.,1.5e-3,NULL,&pars); 
    return result;
  }

  double debugResummation::lum_times_g0expS_matched_der(double N) 
  {
    double h = 0.0001;
    double x = N;
    double y[3];
    double hh;
    for(int i=0; i<3; i++)
    {
      hh = h/pow(2.,i);
      volatile double tmp = x+hh;  // see numerical recipies
      hh = tmp-x;
      y[i] = real( lum_times_g0expS_matched(x+hh)-lum_times_g0expS_matched(x-hh) ) /hh/2.;
    }
    return (y[0] - 6*y[1] + 8*y[2])/3.;
  }
  
  double debugResummation::lum_times_g0expS_matched_der2(double N) 
  {
    double h = 0.0001;
    double x = N;
    double y[3];
    double hh;
    for(int i=0; i<3; i++)
    {
      hh = h/pow(2.,i);
      volatile double tmp = x+hh;  // see numerical recipies
      hh = tmp-x;
      y[i] = real(( lum_times_g0expS_matched(x+hh)+lum_times_g0expS_matched(x-hh)-2.*lum_times_g0expS_matched(x) ) /hh/hh);
    }
    return (y[0] - 6*y[1] + 8*y[2])/3.;
  }
  
  double debugResummation::LumN_der(double N)
  {
    double h = 0.001;
    double x = N;
    double y[3];
    double hh;
    for(int i=0; i<3; i++)
    {
      hh = h/pow(2.,i);
      volatile double tmp = x+hh;  // see numerical recipies
      hh = tmp-x;
      y[i] = ( LumN(x+hh)-LumN(x-hh) ) /hh/2.;
    }
    return (y[0] - 6*y[1] + 8*y[2])/3.;
  }

  double debugResummation::LumN_der2(double N)
  {
    double h = 0.0001;
    double x = N;
    double y[3];
    double hh;
    for(int i=0; i<3; i++)
    {
      hh = h/pow(2.,i);
      volatile double tmp = x+hh;  // see numerical recipies
      hh = tmp-x;
      y[i] = ( LumN(x+hh)+LumN(x-hh)-2.*LumN(x) ) /hh/hh;
    }
    return (y[0] - 6*y[1] + 8*y[2])/3.;
  }

  double debugResummation::derivative_exponent(double N) 
  {
  //  return lum_times_g0expS_matched_der(N)/real(lum_times_g0expS_matched(N)) + LumN_der(N)/real(LuminosityN(N)) - log(tau);
    return  LumN_der( N)/LumN(N) - log(tau);
  }

  double debugResummation::N0_of_tau() 
  {
    double a0 = 1.01;
    double a1 = 5;
    double tolerance = 1e-12;
    int maxsteps = 999;
    int side=0;
    double r, fr, fs = derivative_exponent(a0), ft = derivative_exponent(a1);
    if(fs*ft > 0) {
      cout << "ERROR:  Falsi Method works only if the two initial guesses sorround the root." << endl;
      exit(-35);
    }
    for (int n=0; n<maxsteps; n++) 
    {
      r = (fs*a1 - ft*a0) / (fs - ft);
      if (fabs(a1-a0) < tolerance*fabs(a1+a0)) return r;
      fr = derivative_exponent(r);
      if (fr * ft > 0) 
      {
        a1 = r;
        ft = fr;
        if (side==-1) fs /= 2.;
        side = -1;
      }
      else if (fs * fr > 0) 
      {
        a0 = r;
        fs = fr;
        if (side==+1) ft /= 2;
        side = +1;
      }
      else
        return r;
    }
    cout << "ERROR:  Falsi Method for finding root did not converge." << endl;
    exit(-34);
    return 0;
  }






  double debugResummation::res_xs_of_N0(double m, double sqrt_s,
    double muF_m_ratio, double muR_m_ratio,
    log_order LOGo, order match_order, double *asinput) {
    doRapidityDistr = false;
    //
    tau = m*m/sqrt_s/sqrt_s;
    rapY = 1e300;
    M   = m;
    muF = m*muF_m_ratio;
    muR = m*muR_m_ratio;
    //
    LL_order = LOGo;
    matching_order = match_order;
    //
    if(asinput == NULL) as = alphas(muR); // /50.;
    else as = *asinput; // /50.;
    //cout << "  " << muR << "  "<< as << endl;
    //
    bool redo_cheb = forceredo||!(last_Y==rapY && last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    ComputeCommon(redo_cheb);
    Chebyshev cheb(this);
    //
    //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    //gsl_set_error_handler_off();
    //
    double result = 0.;


    MP_slope = 1.;
    MP_reN0 = MPbeta + 3.;
    double NL = Nlandau();
    if(MP_reN0 >= NL) {
      MP_reN0 = ( MPbeta + NL )/2.;
      if(MP_reN0 >= NL) {
        cout << "ERROR: Landau pole N_L = " << NL
        << " to the left of the rightmost singularity: " << MPbeta << endl;
        exit(-13);
      }
    }
    MPzmin=0.8*tau*(1-tau/2.);
      //if(tau<MPzmin) MPzmin=tau;
    if(tau>0.5) MPzmin=0.3;
    if(redo_cheb) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
    //result = MinimalPrescription()/real(LumN(N0_of_tau()));
    result = real(g0expS_matched(2.70752));
    //cout << real(LuminosityN(100)) << endl;
    //cout << LumN(100) << endl;
    //gsl_set_error_handler (old_handler);
    //
    last_muFrat = muF_m_ratio;
    last_muRrat = muR_m_ratio;
    last_M      = M;
    last_Y      = rapY;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = Hmtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit = false;
    //
  return result;
  }


  dcomplex debugResummation::g0expS_expanded(dcomplex N, order o, log_order LOGo, double m, 
					   double muF_m_ratio, double muR_m_ratio, double *asinput) {
    //tau = m*m/sqrt_s/sqrt_s;
    M   = m;
    muF = m*muF_m_ratio;
    muR = m*muR_m_ratio;
    //
    LL_order = LOGo;
    //
    if(asinput == NULL) as = alphas(muR);
    else as = *asinput;
    //
    ComputeCommon(true);
    //
    return g0expS_expanded(N, o);
  }

  dcomplex debugResummation::muFdep(dcomplex N, order o) {
    using namespace QCD;
    double lf = -2.*log(muFbkp/M);
    double bb1 = b1*b0;
    double bb2 = b2*b0;
    //
    dcomplex C1qg = 0;
    dcomplex C2qg = 0;
    //
    dcomplex C1gg = g0expS_order1(N)/as;
    dcomplex C2gg = g0expS_order2(N)/as/as;
    //
    dcomplex C1SS = 0;
    //
    dcomplex L = -log(N)-EulerGamma;
    dcomplex g0gg = A1*L + gB0/M_PI;
    dcomplex g0gq = 0;
    dcomplex g0qg = 0;
    dcomplex g0qq = 0;//A1/CA*CF*L + 3*CF/4./M_PI;
    //
    dcomplex g1gg = A2*L + gB1/M_PI/M_PI;
    dcomplex g1gq = 0;
    dcomplex g1qg = 0;
    //
    dcomplex g2gg = A3*L + gB2/M_PI/M_PI/M_PI;
    //
    dcomplex FO1gg = 0;
    FO1gg += 4.*(b0 - g0gg)*(-lf/2.);
    //
    dcomplex FO2gg = 0;
    FO2gg += (C1gg*(6*b0 - 4.*g0gg) +4*bb1 - 4.*C1qg*g0qg -4.*g1gg)*(-lf/2.)
      + (3*pow(b0,2) - 5*b0*g0gg + 2.*pow(g0gg,2) + g0gq*g0qg)*pow(-lf,2);
    //
    dcomplex FO3gg = 0;
    FO3gg += (C2gg*(8*b0 - 4.*g0gg) + C1gg*(6*bb1 - 4.*g1gg) - 4.*(-bb2 + C2qg*g0qg + C1qg*g1qg + g2gg))*(-lf/2.)
      + 4.*(-5*bb1*g0gg + 3.*C1qg*g0gg*g0qg + C1SS*pow(g0qg,2) + C1gg*(6*pow(b0,2) - 7*b0*g0gg + 2.*pow(g0gg,2) + g0gq*g0qg) + C1qg*g0qg*g0qq + 
	   b0*(7*bb1 - 7.*C1qg*g0qg - 6.*g1gg) + 4.*g0gg*g1gg + g0qg*g1gq + g0gq*g1qg)*pow(lf/2.,2)
      + (8.*(12*pow(b0,3) - 26*pow(b0,2)*g0gg - 4.*pow(g0gg,3) - 5.*g0gg*g0gq*g0qg + 9*b0*(2.*pow(g0gg,2) + g0gq*g0qg) - g0gq*g0qg*g0qq)*pow(-lf/2.,3))/3.;
    //
    dcomplex res = 0;
    if(o == 1) res = FO1gg*as;
    if(o == 2) res = FO2gg*as*as;
    if(o == 3) res = FO3gg*as*as*as;
    return res;
  }


};






